from pyexpat import model
import socket, numpy as np
from click import command
from tabnanny import verbose
from time import time
from datetime import datetime
import MetaTrader5 as mt5
import pandas as pd
import pytz
from requests import head
from sklearn import preprocessing
from tensorflow.keras.models import load_model
import multiprocessing as multiprocess
import sys
IS_TEST=False
SERVER_ADDRESS="127.0.0.2"
COMM_PORT=9090
SYMBOLS={"EURUSD":9092,"GBPUSD":9093,"NZDUSD":9094,"USDCAD":9095,"USDCHF":9096,
                                             "USDJPY":9097,"XAUUSD":9098}
START_TIMES={}
processes={}
class socketserver:
    def __init__(self, port,modelhelper,symbol_name):
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.port = port
        self.model=modelhelper
        self.conn=0
        self.symbol_name=symbol_name
        self.sock.bind((SERVER_ADDRESS, self.port))
        self.cummdata = ''
        print(f"---->python server initiated successfully:{self.symbol_name}")
        START_TIMES[symbol_name]=str(datetime.now())
    def recvmsg(self):
        r=""
        self.sock.listen(1)
        self.conn, self.addr = self.sock.accept()
        print(f"---->connected to {self.addr}  {self.symbol_name}")
        self.cummdata = ''
        while True:
            data = self.conn.recv(100)
            c=data.decode("utf-8")
            self.cummdata+=c
            if not data:
                break 
            if(self.cummdata=="p"):
                r=self.model.predict()
                self.conn.send(bytes(str(r), "utf-8"))

        log(f"{datetime.now()} : prediction={r}_message={self.cummdata}",self.symbol_name)
        return self.cummdata,r
            
    def __del__(self):
        self.sock.close()
class DataHelper:
    def __init__(self,symbol_name) -> None:
        self.symbol_name=symbol_name
        self.LAST_CLOSE=0
        self.close_scaler = preprocessing.MinMaxScaler()
        #self.close_scaler.fit(pd.read_csv(f"data/walid/{DATA}.csv"
         #          ,names=["time", "open", "high", "low", "close", "volume"])["close"].to_numpy().reshape(-1,1))
        self.timezone = pytz.timezone("Etc/UTC")
        if not mt5.initialize():
            print("---->initialize() failed, error code =",mt5.last_error())
            quit()
        else : print(f"---->mql5 connection established successfully:{self.symbol_name}")
    def __del__(self):
        mt5.shutdown()
        print(f"---->mql5 connection closed successfully:{self.symbol_name}")
    def __fetchData(self):
        if IS_TEST:
            ticks=self.__testTicks()
        else:
            utc_from = datetime.now(self.timezone)
            ticks = mt5.copy_rates_from(self.symbol_name,mt5.TIMEFRAME_M1,utc_from,5000)
        data=pd.DataFrame(data=ticks,columns=ticks.dtype.names)
        data=data.iloc[::-5,:].sort_index(ascending=True)
        self.LAST_CLOSE=data["close"].tail(1).values.item()
        data.drop(labels=["spread","real_volume","tick_volume"],axis=1,inplace=True)
        #--------------------------------------------------------------------------
        data["date"]=data["time"].apply(lambda x:pd.Period(datetime.fromtimestamp(x),"D").dayofyear)
        data["time"]=data["time"].apply(lambda x:datetime.fromtimestamp(x).hour*60+datetime.fromtimestamp(x).minute)
        data["date"] = data["date"].astype('category')
        data["time"] = data["time"].astype('category')
        data=data[["date","time","open","high","low","close"]]
        #--------------------------------------------------------------------------
        self.close_scaler.fit(data["close"].to_numpy().reshape(-1,1))
        #--------------------------------------------------------------------------
        scaler = preprocessing.MinMaxScaler()
        data = scaler.fit_transform(data)
        #--------------------------------------------------------------------------
        return data
    def __testTicks(self):
        utc_from = datetime(2020, 1, 10, tzinfo=self.timezone)
        return  mt5.copy_rates_from(self.symbol_name,mt5.TIMEFRAME_M1,utc_from,5000)
    def makeInputDataset(self):
        data=self.__fetchData()
        return data.reshape((1,1000,6))
class ModelHelper:
    def __init__(self,symbol_name,dataHelper) -> None:
        self.dataHelper=dataHelper
        self.symbol_name=symbol_name
        self.model=load_model(f"./data/walid/model_{symbol_name}.save")
    def predict(self):
        r=0
        try:
            data=self.dataHelper.makeInputDataset()
            r=self.dataHelper.close_scaler.inverse_transform(
                        self.model.predict(data))[0,0]
        except Exception as e:
            log(e,START_TIMES[self.symbol_name],self.symbol_name)
        return r

def log(log,SYMBOL_NAME):
    with open(f'log_{SYMBOL_NAME}.txt', 'a') as f:
        f.writelines(log+"\n")
def startprocess(symbol_name,port_num):
    datahelper=DataHelper(symbol_name)
    modelhelper=ModelHelper(symbol_name,datahelper)
    ss=socketserver(port=port_num,modelhelper=modelhelper,symbol_name=symbol_name)
    while True:
        recvmsg,r=ss.recvmsg()
        if recvmsg=="e":
            break
        elif recvmsg=="p":
            print(f">>>>>>>>predicted value : {r}        {symbol_name}")
    del ss
def start_all():
    processes={}
    try:
        for i in range(1,len(sys.argv)):
            symbol=sys.argv[i]
            p=multiprocess.Process(target=startprocess,
                              args=(symbol,SYMBOLS[symbol],))
            processes[symbol]=p
            p.start()
    except Exception as e:
        log(e,START_TIMES[symbol],symbol)
    return processes
def close_all(processes):
    try:
        for i in range(1,len(sys.argv)):
            symbol=sys.argv[i]
            p=processes[symbol]
            p:multiprocess.Process
            p.terminate()
        print("all processes closed successfully")
    except Exception as e:
        log(e,START_TIMES[symbol],symbol)

if __name__=="__main__":
    command_msg="input your command :"
    while True:
        cmd=input(command_msg)
        if cmd=="start":
            processes=start_all()
        elif cmd=="close":
            close_all(processes)
            processes={}
        elif cmd=="restart":
            close_all(processes)
            processes=start_all()
        elif cmd=="exit":
            print(len(processes))
            close_all(processes)
            processes={}
            break

    """""
    d=DataHelper()
    r=d.fetchData()
    r["time"]=r["time"].apply(lambda x:datetime.fromtimestamp(x))
    print(r.tail(55))
    """""